"""  Module with tools for h5cross
"""

import numpy as np

__all__ = ["add_numpy_statistics", "indentify_next_nested_level"]

# def delete_dict_entries(dict_, keys_to_remove_):
#     """
#         Remove iteratively the desired entries at level 0 if they exist
#         Input:
#             dict_ : dictionary to iterate through
#             keys_to_remove_ : list of type string with entries to be removed

#         Output: None, modifies the input dictionary
#         TO DO: could add option to remove keys at a given level
#             --> need to work on this, or iterative way but careful as might
#                 delete undesired entries perhaps?
#             --> more easily: create another function that calls this function
#                 with dict of certain level! --> YES BETTER, cfr get_desired_field_values
#     """
#     for key in dict_:
#         for ii in keys_to_remove_:
#             if ii in dict_[key]:
#                 del dict_[key][ii]

# def delete_dict_entries_at_level(dict_, keys_to_remove_, level_):
#     Should input either level in a number or a key which will set the level?


# def check_environmental_modules(module_name_list):
#     """ Function checking if desired modules can be imported

#         Input:
#             module_name_list: list of type string containing names of modules to check
#         Output:
#             module_present: list of type boolean listing if module can be imported
#     """

#     module_present=[]
#     for item in module_name_list:
#         try:
#             __import__(item)
#             #print("Module exists")
#             module_present.append(True)
#         except ImportError:
#             #print("Module does not exists")
#             module_present.append(False)
#     return module_present


# def add_numpy_statistics_previous(dict_, flow_, stat_list_= None, skip_list_ = None):
#     """
#         Adds desired (based on numpy library) statistics to a dictionary tree structure

#         Input:
#             :dict_: dictionary to which statistics will be added
#             :flow_: flow field dictionary containing the data to calculate the statistics
#             :stat_list_: list of type string containing statistics to be computed.
#                         Must be keywords known by numpy.
#             :skip_list_: list of type string with dict_ keywords  for which statistics
#                         will not be computed.

#         Output:
#             None, modifies the dict_ entry
#     """
#     if stat_list_ is None:
#         stat_list_ = ["min", "max", "mean", "median", "std"]

#     for key in dict_:
#         if (skip_list_ is None) or (not key in skip_list_):
#             print(key)
#             for keys in dict_[key]:
#                 print("  ", keys)
#                 for item in stat_list_:
#                     try:
#                         dict_[key][keys][item] = str(eval("np."+item)(flow_[key][keys]))
#                     except ValueError:
#                         print("Statistics keyword \"", item, \
#                                 "\" not recognised by numpy for given key:", keys)
#                         #pass
#         else:
#             print("Skip key:", key)




def indentify_next_nested_level(dict_):
    """ Function indentifying if nested object is of type dictionary.
        Loops through the input dictionary keys.

        Input:
            dict_: the dictionary to study
        Output:
            local_dict_: a dictionary with keywords "False" or "True",
                        implying whether for a given key, the nested object
                        is a dictionary or not.
    """
    local_dict_ = {}
    for keys in dict_:
        local_dict_[keys] = False
        if isinstance(dict_[keys], dict):
            local_dict_[keys] = True
    return local_dict_


def compute_numpy_statistics(dict_, array_, stat_list_):
    """ Perform the computation of existing numpy statistics

        Input:
            dict_: dictionary to which statistics will be added
            array_: data array to be used to compute the stats
            stat_list_: list of type string containing statistics to be computed.
                        Must be keywords known by numpy.
        Output: adds stat keys to the input dictionary
    """
    accepted_types = ["float32","float64","int32","int64"]
    # if (dict_["dtype"] in accepted_types) and ("array" in dict_["value"]):

    def _evaluate_numpy_stat(dict_, array_, stat_list_):
        """ Apply the numpy statistics to the given arrays
        """
        for item in stat_list_:
            try:
                dict_[item] = str(eval("np."+item)(array_))
            except NameError: # could be AttributError instead
                print("Statistics keyword \"", item, "\" not recognised by numpy")

    if (dict_["dtype"] in accepted_types) and isinstance(dict_["value"],str):
        _evaluate_numpy_stat(dict_, array_, stat_list_)
        # for item in stat_list_:
        #     try:
        #         dict_[item] = str(eval("np."+item)(array_))
        #     except NameError: # could be AttributError instead
        #         print("Statistics keyword \"", item, "\" not recognised by numpy")
    else:
        try:
            # if single component array with numeric entry let's also add stats
            float(array_)
            _evaluate_numpy_stat(dict_, array_, stat_list_)
        except ValueError:
            pass
        except TypeError:
            pass

def findkeys(node, kv):
    """
        Function allowing to find instances of a key in a nested dictionary.
        
        Code snippet taken from
        https://stackoverflow.com/questions/9807634/find-all-occurrences-of-a-key-in-nested-dictionaries-and-lists
        Credit to arainchi

        Input:
            node: dictionary to iterate through
            kv: keyword to find
        Output:
            generator with result of the search
    """
    if isinstance(node, list):
        for ii in node:
            for item in findkeys(ii, kv):
                yield item
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for jj in node.values():
            # we skip current level and go to next level
            for item in findkeys(jj, kv):
                yield item


def get_desired_field_values(field_, fieldname_):
    """ Function will give the list of values back from the desired field name if it exists

        Input:
            field_:  dictionary of hdf5 object including the field data
            fieldname_: type string, key to find in the field_ dictionary
        Output:
            numpy array of desired field data if it exists
    """
    vals = list(findkeys(field_, fieldname_))
    # assume only one instance in vals -> perhaps need to add option of multiple
    if not vals:
        print("Field \"", fieldname_, "\" does not exist")
        return None
    if len(vals) > 1:
        print("Multiple instances of the desired variable \"", fieldname_, \
                            "\": selected first encounter")
    return vals[0]


def add_numpy_statistics(dict_, flow_, stat_list_= None, skip_list_ = None):
    """
        Adds desired (based on numpy library) statistics to a dictionary tree structure.
        Default stats are "min", "max", "mean", "median" and "std".

        Input:
            :dict_: dictionary to which statistics will be added
            :flow_: flow field dictionary containing the data to calculate the statistics
            :stat_list_: list of type string containing statistics to be computed.
                        Must be keywords known by numpy. (default = None)
            :skip_list_: list of type string with dict_ keywords  for which statistics
                        will not be computed. (default = None)

        Output:
            :None: modifies the dict_ entry
    """
    # define default statistics to compute if no user input
    if stat_list_ is None:
        stat_list_ = ["min", "max", "mean", "median", "std"]

    for key in dict_:
        if (skip_list_ is None) or (not key in skip_list_):
            dict_type_ = indentify_next_nested_level(dict_[key])
            tmp_key = list(dict_type_.keys())[0]
            if not dict_type_[tmp_key]:
                compute_numpy_statistics(dict_[key], flow_[key], stat_list_)
            else:
                add_numpy_statistics(dict_[key], flow_[key], stat_list_, skip_list_)



# def replace_string_in_dict(dict_, key_, string_original_, string_replace_, skip_list_ = None):
#     #TO DO: rethink this, should be level based, so perhaps 2functions cfr delete_dict_entries
#           -> perhaps function that calls given to argument at given level
#     # What if multilevel? could be annoying then eh ...
#     """
#        Replace encounters of a string inside a dictionary.
#     """
#     for key in dict_:
#         if (skip_list_ is None) or (not key in skip_list_):
#             print(key)
#             for keys in dict_[key]:
#                 if key_ in dict_[key][keys]:
#                     if isinstance(dict_[key][keys][key_], str) and \
#                         (string_original_ in dict_[key][keys][key_]):
#                         dict_[key][keys][key_] =  \
#                                  dict_[key][keys][key_].replace(string_original_,string_replace_)
#         else:
#             print("Skip key:", key)



"""h5cross Package"""

# __version__ = "0.0.1.a0"

# from .visit_h5 import *
# from .write_h5 import *
# from .tools import *
# from .view_h5 import *
# from .generate_h5 import *
# from .convert_h5 import *
# from .same_nob import *
# from .h5_checksamefiles import *


""" Functionalities to convert a certain format to hdf5.
    Currently supported:
        - nek5000 (requires "pymech" package)
        - vtk: vtu, pvtu from ALYA solver, vtk unstructured from OpenFOAM (requires "vtk" package)

    TODO: refactor this complete module, splitting functions and rethink the design
"""
import sys
import numpy as np
from .generate_h5 import fill_dict_empty_nested

PYMECH_PRESENT = True
try:
    from pymech.dataset import open_dataset
except ImportError:
    PYMECH_PRESENT = False

VTK_PRESENT = True
try:
    import vtk
except ImportError:
    VTK_PRESENT = False


__all__ = ["convert_nek5000","convert_vtu","convert_vtk"]


def _add_values_to_dict_from_dataset_vtu(dict_, dataset_ = None,
                                            skip_list_ = None, add_mesh_ = True):
    """ Function adding vtk unstructured formatted field variables in a given dictionary
        containing variable names.

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset_: vtk file reader output object
            :skip_list_: optional list of type string with dict_ inputs for which
                        to not add the fields
        Output:
            :None: adds fields to the input dict_
    """

    if dataset_ is None:
        print("Please specify the dataset object")
        sys.exit()
    if skip_list_ is None:
        skip_list_ = []

    # Field variables part
    pointdata = dataset_.GetPointData()
    for variter, varname in enumerate(dict_.keys()):
        if not varname in skip_list_:
            array = pointdata.GetArray(variter)
            # Check for multicomponent 1D arrays
            if array.GetNumberOfComponents() > 1:
                # structure in (X0,Y0,Z0) (X1,Y1,Z1) ...
                num_comp =  array.GetNumberOfComponents()
                num_values = array.GetNumberOfValues()
                tmp_data = np.zeros((int(num_values/num_comp),num_comp))

                for ii in range(int(num_values/num_comp)):
                    for jj in range(num_comp):
                        tmp_data[ii][jj] = array.GetValue(ii * num_comp + jj)

                comp_names = []
                for jj in range(num_comp):
                    comp_names.append('Component'+str(jj))
                fill_dict_empty_nested(dict_, nested_key_ = varname, keys_to_add_ = comp_names)

                for kk, component in enumerate(comp_names):
                    dict_[varname][component] = np.array(tmp_data[:,kk])

            else:
                tmp_data = []
                for ii in range(array.GetNumberOfValues()):
                    tmp_data.append(array.GetValue(ii))
                dict_[varname] = np.array(tmp_data )

    if add_mesh_:
        _add_mesh_values_to_dict_from_dataset_vtu(dict_, dataset_)


def _add_mesh_values_to_dict_from_dataset_vtu(dict_, dataset_ = None):
    """ Function to add mesh points info from pvtu or vtu file to the given dictionary

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset_: vtk file reader output object
        Output:
            :None: adds fields to the input dict_
    """
    # Mesh points
    meshpoints = dataset_.GetPoints().GetData()
    num_comp =  meshpoints.GetNumberOfComponents()
    num_values = meshpoints.GetNumberOfValues()
    tmp_data = np.zeros((int(num_values/num_comp),num_comp))

    for ii in range(int(num_values/num_comp)):
        for jj in range(num_comp):
            tmp_data[ii][jj] = meshpoints.GetValue(ii * num_comp + jj)
    comp_names = []
    for jj in range(num_comp):
        comp_names.append('Component'+str(jj))

    dict_['Mesh'] = {}
    fill_dict_empty_nested(dict_, nested_key_ = 'Mesh', keys_to_add_ = comp_names)
    for kk, component in enumerate(comp_names):
        dict_['Mesh'][component] = np.array(tmp_data[:,kk])


def _get_vtkxml_filetype(filename_, readmax_ = 5):
    """ Function returning the filetype of a vtk xml file.

        Input:
            :filename_: input vtk xml file to check
            :readmax_: int, max number of lines of file to search for the type
        Output:
            :myline: str, file type found
            :Exception raised: in case file type not found
    """

    with open(filename_, 'r') as fin:
        for _ in range(readmax_):
            myline = fin.readline()
            if "type=" in myline:
                return myline.split("type=")[1].split(' ')[0].split('"')[1]

        raise ValueError("VTK xml File type not found, did you specify the correct file type?")


def _add_values_to_dict_from_dataset_nek5000(dict_, dataset_ = None, skip_list_ = None):
    """ Function adding nek5000 formatted field variables in a given dictionary
        containing variable names.

        Input:
            :dict_ : empty nested dictionary to fill
            :dataset_: nek5000 dataset object
            :skip_list_: optional list of type string with dict_ inputs for which
                        to not add the fields
        Output:
            :None: adds fields to the input dict_
    """

    if dataset_ is None:
        print("Please specify the dataset object")
        sys.exit()

    if skip_list_ is None:
        skip_list_ = []
    total_len = dataset_.dims['x'] * dataset_.dims['y'] * dataset_.dims['z']

    for key in list(dict_.keys()):
        if not key in skip_list_:
            dict_[key] = np.reshape(dataset_[key].data, total_len)



def convert_nek5000(file_, skip_vars_ = None):
    """ Main function controlling the conversion of a nek5000 file format
        to dictionary for hdf5 output.

        Input:
            :file_: path to nek5000 file
            :skip_vars_: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)
        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """

    if PYMECH_PRESENT:

        if skip_vars_ is None:
            skip_vars_ = []

        dataset = open_dataset(file_)
        var_names = []
        for item in list(dataset.data_vars):
            if item not in skip_vars_:
                var_names.append(item)

        new_dict = dict()
        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = var_names)
        _add_values_to_dict_from_dataset_nek5000(new_dict, dataset)

        # Add some additional info
        new_dict['Parameters'] = {}
        #print(type(dataset.dims['x']))
        new_dict['Parameters']['Dimensions'] = {'Nx':np.array([dataset.dims['x']], dtype = "int32"),
                                                'Ny':np.array([dataset.dims['y']], dtype = "int32"),
                                                'Nz':np.array([dataset.dims['z']], dtype = "int32")}
        new_dict['Parameters']['time'] = np.array([dataset.coords['time'].data], dtype = "float64")

    else:
        raise ImportError("pymech package not available, "
                            + "please install it for the nek5000 convert feature.")

    return new_dict


def convert_vtu(file_, parallel_ = False, skip_vars_ = None):
    """ Main function controlling the conversion of a vtk xml unstructured file format
        to dictionary for hdf5 output.

        NOTE:   Is currently only based on point data and not cell data.
                No connectivity info is added at the moment.

        Input:
            :file_: path to vtk unstructured file
            :skip_vars_: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)
            :parallel_: boolean, specification if we are dealing with pvtu of normal vtu file

        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """

    if VTK_PRESENT:
        if skip_vars_ is None:
            skip_vars_ = []

        # Create reader object
        filetype = _get_vtkxml_filetype(file_)
        if parallel_:
            assert filetype == 'PUnstructuredGrid', "VTK file type does not coincide with user input"
            reader = vtk.vtkXMLPUnstructuredGridReader()
        else:
            assert filetype == 'UnstructuredGrid', "VTK file type does not coincide with user input"
            reader = vtk.vtkXMLUnstructuredGridReader()

        # Read vtk file
        reader.SetFileName(file_)
        reader.Update()

        # Obtain variable names
        vtkdata = reader.GetOutput()
        pointdata = vtkdata.GetPointData()
        vars_num = reader.GetNumberOfPointArrays()
        vars_name = []
        for variter in range(vars_num):
            vars_name.append(pointdata.GetArrayName(variter).split()[0])

        # Create dictionary with structure
        new_dict = dict()
        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = vars_name)

        # Fill dictionary with data
        _add_values_to_dict_from_dataset_vtu(new_dict, vtkdata)
    else:
        raise ImportError("vtk package not available,"
                        + "please install it for the vtk convert features.")

    return new_dict


def convert_vtk(file_, skip_vars_ = None):
    """ Main function controlling the conversion of a vtk file format
        to dictionary for hdf5 output.

        NOTE:   Is currently only based on point data and not cell data.
                No connectivity info is added at the moment.
                Currently only for vtk unstructured.

        Input:
            :file_: path to vtk unstructured file
            :skip_vars_: optional list of type string containing fields we wish to
                not consider transferring to the hdf5 output
                (default = None)

        Output:
            :new_dict: python dictionary format structured for hdf5 write output
    """

    if VTK_PRESENT:
        if skip_vars_ is None:
            skip_vars_ = []

        # Select reader
        reader = vtk.vtkUnstructuredGridReader()

        # Read vtk file
        reader.SetFileName(file_)
        reader.Update()

        # Obtain variable names
        vtkdata = reader.GetOutput()
        pointdata = vtkdata.GetPointData()
        vars_num = pointdata.GetNumberOfArrays()

        vars_name = []
        for variter in range(vars_num):
            vars_name.append(pointdata.GetArrayName(variter).split()[0])

        # Create dictionary with structure
        new_dict = dict()
        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = vars_name)

        # Fill dictionary with data
        _add_values_to_dict_from_dataset_vtu(new_dict, vtkdata)
    else:
        raise ImportError("vtk package not available,"
                        + "please install it for the vtk convert features.")

    return new_dict



from pymech.neksuite import readnek
from pymech.dataset import open_dataset

from generate_h5 import fill_dict_empty_nested
from write_h5 import write_h5
import numpy as np
import sys
import vtk

# https://nek5000.github.io/NekDoc/problem_setup/case_files.html#restart-output-files-f

#------------------------------------------------------------------#
def fill_dict_empty_nested_from_listkeys(dict_, keys_to_add_ = None):
    """ Function calls fill_dict_empty_nested() for each key in present dictionary
    """
    if keys_to_add_ is None:
        print("Please specify dictionary entries to add")
        sys.exit()

    for item in list(dict_.keys()):
        fill_dict_empty_nested(dict_, nested_key_ = item, keys_to_add_ = keys_to_add_)


def add_values_to_dict_from_dataset(dict_, dataset_ = None, skip_list_ = None):
    """ 

        Input:
            :dict_ : empty nested dictionary to fill

    """

    if dataset_ is None:
        print("Please specify the dataset object")
        sys.exit()

    if skip_list_ is None:
        skip_list_ = []
    total_len = dataset_.dims['x'] * dataset_.dims['y'] * dataset_.dims['z']

    for key in list(dict_.keys()):
        if not key in skip_list_:
            dict_[key] = np.reshape(dataset_[key].data, total_len)


#------------------------------------------------------------------#

flag_nek5000 = 0
if flag_nek5000:
    file_loc = "/Users/hoste/tools/coec_wp6/DATA_FORMATS/Nek5000/"
    file_name = "premix0.f00001"
    readfile = readnek(file_loc + file_name) 

    dataset = open_dataset(file_loc + file_name)

    total_len = dataset.dims['x'] * dataset.dims['y'] * dataset.dims['z']
    var_names = list(dataset.data_vars)
    new_dict = dict()
    fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = var_names)
    # Variables are listed in (z,y,x) order
    new_temp = np.reshape(dataset['temperature'].data, total_len)

    add_values_to_dict_from_dataset(new_dict, dataset)

#    write_h5(new_dict, save_name_ = 'dump', save_path_='./')


flag_vtk = 1
if flag_vtk:
    # file_loc = "/Users/hoste/JJ_TAU/Cheng_1_1stO_coarse/BK_reac_2ndO_profEilmer_Gerlinger_ctu_2/vtk/"
    # file_name = "solution.pval.922847_hexa.vtu"

    '''
        Useful paper perhaps: Simple visualizations of unstructured grids with VTK
    '''

    # Individual vtu test
    flag_test1 = 0
    if flag_test1:
        file_loc = "/Users/hoste/tools/coec_wp6/DATA_FORMATS/ALYA/ALYA_OUTPUT/VTK_FORMAT/non_prem_box_VarVar_vectorized_00000000/"
        file_name = "non_prem_box_VarVar_vectorized_00000000_0.vtu"
        ''' File info: VTKFile type="UnstructuredGrid" 
                                version="0.1" 
                                byte_order="LittleEndian" 
                                header_type="UInt32" 
                                compressor="vtkZLibDataCompressor">
        '''
        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(file_loc + file_name)
        reader.Update()

        print("Nb of cells = ", reader.GetNumberOfCells())
        reader.GetNumberOfPoints()

        vars_num = reader.GetNumberOfPointArrays() 
        vtkdata = reader.GetOutput() # this is the actual reading output

        # So either one of both if not mistaken, what if vectors? Need to see different possibilities
        pointData = vtkdata.GetPointData()
        vectorData = pointData.GetVectors()

        # Could be pretty interesting
        # https://blog.kitware.com/hdf5-reader-and-writer-for-unstructured-grids/

        new_dict = dict()
        vars_name = []
        for variter in range(vars_num):
            vars_name.append(pointData.GetArrayName(variter).split()[0])

        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = vars_name)

    #    arr.GetNumberOfValues()
        for variter, varname in enumerate(vars_name):
            array = pointData.GetArray(variter)
            tmp_data = []
            for ii in range(array.GetNumberOfValues()):
                tmp_data.append(array.GetValue(ii))
            new_dict[varname] = np.array(tmp_data )

        #write_h5(new_dict, save_name_ = 'vtkdump', save_path_='./')
        #GetNumberOfValues() 

    # with open(file_loc+file_name, 'r') as fin:
    #     readmax = 5
    #     for line in range(readmax):
    #         if "type=" in fin.readline():
    #             print("mytype")

    #NOTE!!
    # SO WOULD BE ADDING A COMPONENT!
    #<PDataArray type="Float64" Name="VELOC" NumberOfComponents="3"/>
    # ALSO I THINK POINT info is also there, for mesh in float32 type

   # Individual pvtu test
    flag_test2 = 0
    if flag_test2:
        ''' Possibly https://github.com/JiaweiZhuang/vtk_experiments/blob/master/vtki_correct_pvtu.ipynb
            https://github.com/mmagnuski/vtki/tree/master/vtki
            But adds even another depedency, would avoid that

            (maybe but less interesting https://github.com/pearu/pyvtk)

        Header:
            <VTKFile type="PUnstructuredGrid"
                            version="0.1" 
                            byte_order="LittleEndian"
                            header_type="UInt32"
                            compressor="vtkZLibDataCompressor">
            <PUnstructuredGrid GhostLevel="0">
            <PPointData>
            <PDataArray type="Float64" Name="CON01"/>
            <PDataArray type="Float64" Name="CON02"/>
        '''

        file_loc = "/Users/hoste/tools/coec_wp6/DATA_FORMATS/ALYA/ALYA_OUTPUT/VTK_FORMAT/"
        file_name = "non_prem_box_VarVar_vectorized_00000000.pvtu"

        # MAYBE THE GHOSTCELL PART GOT AUTOMATICALLY DEALT WITH By THE READER- > would make sense

        # Use XMLP
        reader = vtk.vtkXMLPUnstructuredGridReader()
        reader.SetFileName(file_loc + file_name)
        reader.Update()

        print("Nb of pieces (individual vtu files) = ", reader.GetNumberOfPieces())
        # Reader seems to join all individuals alreadye AND probable 

        vars_num = reader.GetNumberOfPointArrays() 
        vtkdata = reader.GetOutput()
        pointData = vtkdata.GetPointData()

        new_dict = dict()
        vars_name = []
        for variter in range(vars_num):
            vars_name.append(pointData.GetArrayName(variter).split()[0])

        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = vars_name)

        for variter, varname in enumerate(vars_name):
            array = pointData.GetArray(variter)
            if array.GetNumberOfComponents() > 1:
                print("Nb of components = ", array.GetNumberOfComponents())
                print("Size of array = ", array.GetSize())
                #TODO: NEED TO KNOW IF ALL X, Y, Z comp or (X0,Y0,Z0) (X1,Y1,Z1) ...
                # opening in paraview appears to be the second
                tmp_data = np.zeros((int(array.GetNumberOfValues()/array.GetNumberOfComponents()),
                                     array.GetNumberOfComponents()))
                
                for ii in range(int(array.GetNumberOfValues()/array.GetNumberOfComponents())):
                    for jj in range(array.GetNumberOfComponents()):
                        tmp_data[ii][jj] = array.GetValue(ii * array.GetNumberOfComponents() + jj)

                comp_names = []
                for jj in range(array.GetNumberOfComponents()):
                    comp_names.append('Component'+str(jj))
                fill_dict_empty_nested(new_dict, nested_key_ = varname, keys_to_add_ = comp_names)
                
                for kk, component in enumerate(comp_names):
                    new_dict[varname][component] = np.array(tmp_data[:,kk])

            else:
                tmp_data = []
                for ii in range(array.GetNumberOfValues()):
                    tmp_data.append(array.GetValue(ii))
                new_dict[varname] = np.array(tmp_data )

        #STILL NEED TO ADD POINTS!!
        meshPoints = vtkdata.GetPoints().GetData()
       

        num_comp =  meshPoints.GetNumberOfComponents()
        num_values = meshPoints.GetNumberOfValues()
        tmp_data = np.zeros((int(num_values/num_comp),num_comp))
              
        for ii in range(int(num_values/num_comp)):
            for jj in range(num_comp):
                tmp_data[ii][jj] = meshPoints.GetValue(ii * num_comp + jj)
        comp_names = []
        for jj in range(num_comp):
            comp_names.append('Component'+str(jj))

        new_dict['Mesh'] = {}
        fill_dict_empty_nested(new_dict, nested_key_ = 'Mesh', keys_to_add_ = comp_names)
        for kk, component in enumerate(comp_names):
            new_dict['Mesh'][component] = np.array(tmp_data[:,kk])
        write_h5(new_dict, save_name_ = 'vtkparallel', save_path_='./')


    flag_test3 = 1
    if flag_test3:
        file_loc = "/Users/hoste/tools/coec_wp6/DATA_FORMATS/OpenFoam/VTK/"
        file_name = "001_fireII_NR_s_46115.vtk"
        '''
            Possible interest:
            https://discourse.vtk.org/t/python-reading-calculate-write-out/1138/4
            vtk pyvista possible interest

            meshio https://github.com/nschloe/meshio
        '''

        #reader = vtk.vtkDataSetReader()
        reader = vtk.vtkUnstructuredGridReader()
        # #reader = vtk.vtkStructuredPointsReader()
        
        reader.SetFileName(file_loc + file_name)
        # reader.ReadAllVectorsOn()
        # reader.ReadAllScalarsOn() #--> so probably needed to get the vars
        reader.Update()
        vtkdata = reader.GetOutput()
        
        pointData = vtkdata.GetPointData()
        vars_num = pointData.GetNumberOfArrays()

        new_dict = dict()
        vars_name = []
        for variter in range(vars_num):
            vars_name.append(pointData.GetArrayName(variter).split()[0])

        fill_dict_empty_nested(new_dict, nested_key_ = None, keys_to_add_ = vars_name)


        for variter, varname in enumerate(vars_name):
            array = pointData.GetArray(variter)
            if array.GetNumberOfComponents() > 1:
                print("Nb of components = ", array.GetNumberOfComponents())
                print("Size of array = ", array.GetSize())
                #TODO: NEED TO KNOW IF ALL X, Y, Z comp or (X0,Y0,Z0) (X1,Y1,Z1) ...
                # opening in paraview appears to be the second
                tmp_data = np.zeros((int(array.GetNumberOfValues()/array.GetNumberOfComponents()),
                                     array.GetNumberOfComponents()))
                
                for ii in range(int(array.GetNumberOfValues()/array.GetNumberOfComponents())):
                    for jj in range(array.GetNumberOfComponents()):
                        tmp_data[ii][jj] = array.GetValue(ii * array.GetNumberOfComponents() + jj)

                comp_names = []
                for jj in range(array.GetNumberOfComponents()):
                    comp_names.append('Component'+str(jj))
                fill_dict_empty_nested(new_dict, nested_key_ = varname, keys_to_add_ = comp_names)
                
                for kk, component in enumerate(comp_names):
                    new_dict[varname][component] = np.array(tmp_data[:,kk])

            else:
                tmp_data = []
                for ii in range(array.GetNumberOfValues()):
                    tmp_data.append(array.GetValue(ii))
                new_dict[varname] = np.array(tmp_data )



        #STILL NEED TO ADD POINTS!!
        meshPoints = vtkdata.GetPoints().GetData()
       

        num_comp =  meshPoints.GetNumberOfComponents()
        num_values = meshPoints.GetNumberOfValues()
        tmp_data = np.zeros((int(num_values/num_comp),num_comp))
              
        for ii in range(int(num_values/num_comp)):
            for jj in range(num_comp):
                tmp_data[ii][jj] = meshPoints.GetValue(ii * num_comp + jj)
        comp_names = []
        for jj in range(num_comp):
            comp_names.append('Component'+str(jj))

        new_dict['Mesh'] = {}
        fill_dict_empty_nested(new_dict, nested_key_ = 'Mesh', keys_to_add_ = comp_names)
        for kk, component in enumerate(comp_names):
            new_dict['Mesh'][component] = np.array(tmp_data[:,kk])
        write_h5(new_dict, save_name_ = 'vtkparallel', save_path_='./')
        


""" Functionalities to generate random nested dictionaries.
    Currently implemented for either 1 or 2 nested levels chosen at random.
"""
'''
    Works currently only for a max of two nested levels!!
    Would need to see what to do for more.
    It is highly probably due to the possible issue of
    base_keys vs variables at some point in the tree.
'''

import sys
import numpy as np

__all__ = ["generate_random_nested_dictionary"]

# variable list of tuples with format  ('name', min_val, max_val)
variables = [('temperature',300.0,1000.0),
                ('pressure',1e5,4e5),
                ('density', 0.08,1.98),
                ('scalar dissipation rate',0.0,2400.0),
                ('mixture fraction',0.0,1.0),
                ('O2 mass fraction',0.0,1.0),
                ('H2 mass fraction',0.0,1.0),
                ('H2O mass fraction',0.0,1.0),
                ('CO mass fraction',0.0,1.0),
                ('N2 mass fraction',0.0,1.0),
                ('CO2 mass fraction',0.0,1.0),
                ('O mass fraction',0.0,1.0),
                ('C mass fraction',0.0,1.0),
                ('total enthalpy',-3e6,3e6),
                ('velocity magnitude',20.0,790.0),
                ('Mach number',0.0,2.0),
                ('viscosity',8e-4,2e-5),
             ]

base_keys = {'Additionals', 'GaseousPhase', 'Special', 'Extras', 'New', 'Magic'}


def check_randomness(num_to_check, array_to_check):
    """ Function that checks if there is a minimal occurance of 1 for each value in given range.
        If it's not the case, one of the max occurences will be replaced by the minumum value.

        Input:
            :num_to_check: integer, the highest num in range to check: range = [0,num_to_check]
            :array_to_check: array of int with values to check
        Output
            :new_array: = array_to_check if each value in range at least one time in array_to_check
                        if above statement not correct, will replace one instance of max occurance
                        by missing one
    """
    new_array = array_to_check
    for item in np.arange(num_to_check):
        if item in new_array:
            continue
        count = np.argmax(np.bincount(new_array))
        index_replace = np.where(new_array == count)[0][0]
        new_array[index_replace] = item
        #print("Replaced value:", count, ' at index ',
        # index_replace, ' by missing value ', item)
    return new_array

def fill_dict_empty_nested(dict_, nested_key_ = None, keys_to_add_ = None):
    """ Function to add nested levels in a dictionary.
        Empty nested dictionaries are added for given keywords to be addded.

        Input:
            dict_: dictionary to which nested levels will be added
            nested_key_: (optional) type string, sets the dictionary level
                         to which to add keys (default = None)
            keys_to_add_ = keyword entries to add to dictionary.
        Output:
            None, modifies input dict_
    """
    if keys_to_add_ is None:
        print("Please specify dictionary entries to add")
        sys.exit()

    if nested_key_ is None:
        for item in keys_to_add_:
            if not dict_ == '':
                dict_[item] = {}
            else:
                dict_ = {item : ""}
    else:
        for item in keys_to_add_:
            if not dict_[nested_key_] == '':
                dict_[nested_key_][item] = {}
            else:
                dict_[nested_key_] = {item : ""}


def fill_dict_with_random_values(dict_, vars_info_, nvals = 20):
    """ Function that will generate the data to be added to the dictionary.

        Input:
            dict_: dictionary where to add the data arrays
            vars_info_: list containing in order
                        1) the dictionary key of where to add the value
                        2) min allowed value of data
                        3) max allowed value of data
        Output:
            None, adds a numpy array of data to the input dictionary
    """
    for item in vars_info_:
        key = item[0]
        var_min = float(item[1])
        var_max = float(item[2])
        #print(key, var_min, var_max)
        values = np.random.uniform(low = var_min, high = var_max, size = nvals)
        dict_[key] = values


def select_elements_from_list_from_value(check_list, select_list, target_value):
    """ Function that selects elements from a list based on a target value.
        Use of numpy where functionality.

        Input:
            check_list: list of values to browse through and evaluate our target value
            select_list: list to select from based on evaluation (could be same as check list)
            target_value: single value to find in check_list
        Output:
            select_elements: list of elements corresponding to target condition

    """
    tmp = np.where(check_list == target_value)[0]
    select_elements = [list(select_list)[index] for index in tmp]
    return select_elements

def select_next_nested_dict_level(dict_):
    """ Function accessing the next nested level of a dictionary.
        Reference to the original input dictionary is kept, i.e. no deepcopy returned.

        Input:
            dict_: nested dictionary from which to select next nested level
        Output:
            new_dict: dictionary starting at the next nested level of the input dict
    """
    new_dict = dict()
    for index, key in enumerate(list(dict_.keys())):
        if index == 0:
            #print(key)
            new_dict = dict_[key]
        else:
            new_dict[key] = dict_[key]
    return new_dict


def get_random_nested_dict(num_nested_levels_,
                            base_keys_, base_reshuffle_,
                            var_keys_, var_reshuffle_,
                            array_length_ = 20):
    """ Function generating a random nested dictionary with data points.

        Input:
            num_nested_levels_: int, sets the number of nested levels
                                (must be either 1 or 2 in current implementation)
            base_keys_: list of type string, contains the dictionary entries for
                        the different levels
            base_reshuffle_: list of type int, random assignment of base_keys_ to the
                            different levels
            var_keys_: list of type tuples, the variables for which to generate data
                        and add to the dictionary
            var_reshuffle_: list of type int, random assignment of var_keys_ to the
                            different levels
            array_length_: (optional), type int, controls the length of the generated
                            data for each entry of var_keys_
        Output:
            mydict: random nested dictionary

    """
    mydict = dict()
    tmp_dict = dict()
    for ii in np.arange(num_nested_levels_):
        # Select our base keys to add to the dict
        select_keys = select_elements_from_list_from_value(base_reshuffle_, base_keys_, ii)
        num_keys = len(select_keys)

        # Select the variables to add to the dict
        select_vars = select_elements_from_list_from_value(var_reshuffle_, var_keys_, ii)

        # Distribute our variables over the number of base keys through a random shuffle
        local_shuffle = np.random.randint(num_keys, size = len(select_vars))
        local_shuffle = check_randomness(num_keys, local_shuffle)

        if ii == 0:
            fill_dict_empty_nested(mydict, None, select_keys)
            for jj in range(len(select_keys)):
                vars_to_add = select_elements_from_list_from_value(local_shuffle, select_vars, jj)

                fill_dict_empty_nested(mydict, select_keys[jj], np.array(vars_to_add)[:,0])
                fill_dict_with_random_values(mydict[select_keys[jj]],
                                                np.array(vars_to_add),
                                                nvals = array_length_)
                tmp_dict = mydict
        else:
            # We need to distribute the next nested level (base keys) on the existing
            # last nested level base keys branches
            next_base_shuffle = np.random.randint(len(tmp_dict.keys()), size = len(select_keys))
            next_base_shuffle = check_randomness(len(tmp_dict.keys()), next_base_shuffle)
            for kk in range(len(tmp_dict.keys())):
                base_to_add = select_elements_from_list_from_value(next_base_shuffle,
                                                                     select_keys, kk)
                nested_level = list(tmp_dict.keys())[kk]
                fill_dict_empty_nested(tmp_dict[nested_level], None, base_to_add)

            for jj in range(len(select_keys)):
                vars_to_add = select_elements_from_list_from_value(local_shuffle, select_vars, jj)
                #print(vars_to_add)
                nested_level = list(tmp_dict.keys())[next_base_shuffle[jj]]

                fill_dict_empty_nested(tmp_dict[nested_level],
                                        select_keys[jj],
                                        np.array(vars_to_add)[:,0])
                fill_dict_with_random_values(tmp_dict[nested_level][select_keys[jj]],
                                                        np.array(vars_to_add),
                                                        nvals= array_length_)

            tmp_dict = select_next_nested_dict_level(tmp_dict)

    return mydict



def generate_random_nested_dictionary(array_length = 20):
    """ Function controlling the generation of a random nested dictionary.
        The randomness will generate a dictionary with either 1 or 2 nested levels.

        Input:
            array_length: (optional), sets the length of the data that is generated
        Output:
            mydict: nested dictionary
    """
    # TO DO: add exception if num_nested_levels = 0 -> then would need to just generate a
    # list of all values instead and do not consider the base keys then
    #       ->perhaps generare a list of vars instead in a seperate function

    num_nested_levels = np.random.randint(1,3,size=1)

    var_reshuffle = np.random.randint(num_nested_levels, size = len(variables))
    var_reshuffle = check_randomness(num_nested_levels, var_reshuffle)

    base_reshuffle = np.random.randint(num_nested_levels, size = len(base_keys))
    base_reshuffle = check_randomness(num_nested_levels, base_reshuffle)

    # TO DO: add Parameters with some random fill, at level 0 : like nnode and min vol and other
    # -->  lik a standard to add

    mydict = get_random_nested_dict(num_nested_levels, base_keys,
                                    base_reshuffle, variables,
                                    var_reshuffle, array_length_= array_length)

    return mydict



""" Module with functionalities related to the visual aspects of h5cross
"""
import subprocess
import sys
import numpy as np
from .visit_h5 import get_h5_structure, get_h5_field
from .write_h5 import write_dict_to_yaml
from .tools import add_numpy_statistics, get_desired_field_values

__all__ = ["compare_h5","visual_h5","compare_scatter_h5"]


NOBVISUAL_PRESENT = True
try:
    import nobvisual
except ImportError:
    NOBVISUAL_PRESENT = False

MATPLOTLIB_PRESENT = True
SEABORN_PRESENT = True
try:
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib.colors import Normalize
    try:
        import seaborn as sns
    except ImportError:
        SEABORN_PRESENT = False
except ImportError:
    MATPLOTLIB_PRESENT = False

SCIPY_PRESENT = True
try:
    from scipy.interpolate import interpn
except ImportError:
    SCIPY_PRESENT = False


def compare_h5(file1, file2, add_stats = False):
    """ Function allowing the comparison of two hdf5 files

        Input:
            :file1: (string), name of first file for comparison
            :file2: (string), name of second file for comparison
            :add_stats: (boolean), option to compute statistics on both input files
        Output:
            :None: interactive comparative view
    """

    if NOBVISUAL_PRESENT:
        dict1 = get_h5_structure(file1)
        dict2 = get_h5_structure(file2)

        # TO DO: write separate functionalities for this??
        if add_stats:
            flow1 = get_h5_field(file1)
            add_numpy_statistics(dict1, flow1)
            del flow1
            flow2 = get_h5_field(file2)
            add_numpy_statistics(dict2, flow2)
            del flow2

        write_dict_to_yaml(dict1, "tmp1", setindent = 2)
        write_dict_to_yaml(dict2, "tmp2", setindent = 2)

        _visual_compare_yaml("tmp1.yml","tmp2.yml")

        # TODO: replace by os.path or shutil removal function 
        subprocess.call('rm '+ "tmp1.yml" + ' ' + "tmp2.yml", shell=True)
    else:
        raise ImportError("nobvisual package not available, please install it for this feature")


def _visual_compare_yaml(yaml1,yaml2):
    """ Function calling nobvisual for visual comparison of yaml files

        Input:
            :yaml1: name of first yaml file
            :yaml2: name of second yaml file
        Output:
            :None: interactive view
    """
    if NOBVISUAL_PRESENT:
        subprocess.call('nobvisual cmpfile ' + yaml1 + ' ' + yaml2, shell=True)
    else:
        raise ImportError("nobvisual package not available, please install it for this feature")


def visual_h5(filename):
    """ Function calling nobvisual for visual view of hdf5 file

        Input:
            :filename: name of hdf5 file
        Output:
            :None: interactive view
    """

    if NOBVISUAL_PRESENT:
        dict_ = get_h5_structure(filename)
        write_dict_to_yaml(dict_, "tmp", setindent = 2)
        _call_nobvisual_treefile_on_h5('tmp.yml', remove_file = True)
    else:
        raise ImportError("nobvisual package not available, please install it for this feature")

def _call_nobvisual_treefile_on_h5(filename, remove_file = True):
    """ Separate function call to nobvisual treefile

        Input:
            :filename: (string), yaml file to read with nobvisual
            :remove_file: (boolean), whether to delete yaml file upon completion
    """

    subprocess.call('nobvisual treefile '+ filename, shell=True)
    subprocess.call('rm '+ filename , shell=True)

def compare_scatter_h5(file1, file2, variable_list,
                         flag_save = False, save_name = None,
                         flag_show = True, flag_seaborn = True,
                         flag_xy_line = False):
    """ Function that controls the generation of scatterplots comparing the same
        variable from two hdf5 files

        Input:
            :file1: (string), name of first file for comparison
            :file2: (string), name of second file for comparison
            :variable_list: list of type string containing keywords of the
                            variables to select for comparison. The same variable
                            will be selected from each file.
            Optional Arguments:
                :flag_save: boolean controlling if the plots get saved (default = False)
                :save_name: (string), base name to be used in saving the plots (default = None)
                :flag_show: boolean setting if the plot gets showed or not (default = True)
                :flag_seaborn: boolean controlling whether seaborn is used or not (default = True)
        Output:
            :None: interactive view. If flag_save = True, a png image of the scatter plots gets saved.
    """

    if MATPLOTLIB_PRESENT:
        flow1 = get_h5_field(file1)
        flow2 = get_h5_field(file2)
        for item in variable_list:
            array1 = get_desired_field_values(flow1, item)
            array2 = get_desired_field_values(flow2, item)
            if (array1 is not None) and (array2 is not None):
                if len(array1) == len(array2):
                    if (SEABORN_PRESENT) and (flag_seaborn is True):
                        seaborn_simple_scatter(array1, array2, [item,item],
                                flag_save_ = flag_save, flag_show_ = flag_show,
                                save_name_ = save_name, flag_xy_line = flag_xy_line)
                    else:
                        matplotlib_simple_scatter(array1, array2, [item,item],
                                flag_save_ = flag_save, flag_show_ = flag_show,
                                save_name_ = save_name, flag_xy_line = flag_xy_line)
                else:
                    print("Warning: data arrays have different length, plotting distribution instead")
                    #if modules_present[1] is not False:
                    #    seaborn_simple_distribution(array1, array2, [item,item],
                    #            flag_save_ = flag_save, flag_show_ = flag_show, save_name_ = save_name)
                    #else:
                    matplotlib_simple_distribution(array1, array2, [item,item],
                                flag_save_ = flag_save, flag_show_ = flag_show, save_name_ = save_name)
    else:
        raise ImportError("matplotlib package not available, please install it for this feature")


def seaborn_simple_scatter(array1, array2, axes_labels,
                            flag_save_ = False, save_name_ = None,
                            flag_show_ = True, flag_xy_line = False):
    """ Function to generate a scatter plot of two data arrays of same length with seaborn

        Input:
            array1: numpy array of first data set
            array2: numpy array of second data set
            axes_labels: list of string of length 2, used to specify the labels of output axis
            Optional Arguments:
                flag_save_: boolean controlling if the plots get saved (default = False)
                save_name_: (string), base name to be used in saving the plots (default = None)
                flag_show_: boolean setting if the plot gets showed or not (default = True)

        Output:
            Default = None, interactive view
            If flag_save = True then outputs a png image
    """

    sns.set_theme()

    if SCIPY_PRESENT:
        # Get the density map
        zarr = _density_scatter(array1, array2, only_cmap = True)

        # Sort the data for better color map rendering (see link)
        idx = zarr.argsort()
        array1, array2, zarr = array1[idx], array2[idx], zarr[idx]

    data_plot = {"field1": array1, "field2": array2}
    plot = sns.jointplot(data = data_plot, x = "field1", y = "field2", kind = "scatter",
                            marginal_ticks= True,
                            marginal_kws=dict(stat="density") )
    if flag_xy_line:
        xmin = np.min(np.append(array1,array2))
        xmax = np.max(np.append(array1,array2))
        plot.ax_joint.plot([xmin,xmax], [xmin,xmax], 'k-', linewidth = 2)

    if SCIPY_PRESENT:
        plot.plot_joint(plt.scatter, c=zarr, cmap='viridis')
    plot.set_axis_labels(axes_labels[0] + ' ' + "field 1" ,axes_labels[1] +' ' + "field 2")

    plt.tight_layout()
    if flag_save_:
        if not save_name_:
            base_name = "scatter_"
        else:
            base_name = save_name_+"_"
        save_name = base_name + axes_labels[0] + '_' + axes_labels[1]
        plt.savefig(save_name +'.png')
    if flag_show_:
        plt.show()


def seaborn_simple_distribution(array1, array2, axes_labels,
                            flag_save_ = False, save_name_ = None, flag_show_ = True):
    """ Function to generate histograms of two data arrays of different length with seaborn
        NOT YET FUNCTIONAL

        Input:
            array1: numpy array of first data set
            array2: numpy array of second data set
            axes_labels: list of string of length 2, used to specify the labels of output axis
            Optional Arguments:
                flag_save_: boolean controlling if the plots get saved (default = False)
                save_name_: (string), base name to be used in saving the plots (default = None)
                flag_show_: boolean setting if the plot gets showed or not (default = True)

        Output:
            Default = None, interactive view
            If flag_save = True then outputs a png image
    """
    import seaborn as sns
    import matplotlib.pyplot as plt
    #TO DO: create a dictionary for default input items instead of list
    sns.set_theme()
    # Will be deprecated soon, must update data structure
    # https://seaborn.pydata.org/tutorial/data_structure.html
    plot = sns.displot(array1)
    plot.displot(array2)
    plot.set_axis_labels(axes_labels[0], "count")
    plt.tight_layout()
    if flag_save_:
        if not save_name_:
            base_name = "dist_"
        else:
            base_name = save_name_+"_"
        save_name = base_name + axes_labels[0] + '_' + axes_labels[1]
        plt.savefig(save_name +'.png')
    if flag_show_:
        plt.show()


def _density_scatter( xarr , yarr, ax = None, sort = True, bins = 24, only_cmap = False, **kwargs ):
    """
        Scatter plot colored by 2d histogram with matplotlib

        Code snippet taken from (adapted naming convention)
        https://stackoverflow.com/questions/20105364/how-can-i-make-a-scatter-plot-colored-by-density-in-matplotlib
        Credits to: Guillaume

        Input:
            arr1: numpy array of first data set
            arr2: numpy array of second data set
            ax: default = None, matplotlib Axes object
            sort: boolean = True, sorting color map values before plotting
            bins: default = 24, integer specifying number of bins in both directions of 2D histogram
            only_cmap: default = False, if True the function only returns the density map

        Output:
            ax: matplotlib Axes object
    """

    if SCIPY_PRESENT:
        data , xedge, yedge = np.histogram2d( xarr, yarr, bins = bins, density = True )

        zarr = interpn( ( 0.5*(xedge[1:] + xedge[:-1]),0.5*(yedge[1:]+yedge[:-1]) ),
                    data ,
                    np.vstack([xarr,yarr]).T ,
                    method = "splinef2d", bounds_error = False)

        # To be sure to plot all data
        zarr[np.where(np.isnan(zarr))] = 0.0

        # If we're only interested in the density color map
        if only_cmap:
            return zarr

        # Sort the points by density, so that the densest points are plotted last
        if sort :
            idx = zarr.argsort()
            xarr, yarr, zarr = xarr[idx], yarr[idx], zarr[idx]

        if ax is None :
            fig , ax = plt.subplots()

        ax.scatter( xarr, yarr, c=zarr, **kwargs )

        if MATPLOTLIB_PRESENT:
            norm = Normalize(vmin = np.min(zarr), vmax = np.max(zarr))
            cbar = fig.colorbar(cm.ScalarMappable(norm = norm), ax=ax)
            cbar.ax.set_ylabel('Density')
        else:
            raise ImportError("matplotlib package not available, please install it for the density colored scatter feature")
    else:
        raise ImportError("scipy package not available, please install it for the density colored scatter feature with matplotlib")

def matplotlib_simple_scatter(array1, array2, axes_labels,
                                flag_save_ = False, save_name_ = None,
                                flag_show_ = True, flag_xy_line = False):
    """ Function to generate a scatter plot of two data arrays of same length with matplotlib.
        If scipy is available a density scatter plot will be generated.

        Input:
            array1: numpy array of first data set
            array2: numpy array of second data set
            axes_labels: list of string of length 2, used to specify the labels of output axis
            Optional Arguments:
                flag_save_: boolean controlling if the plots get saved (default = False)
                save_name_: (string), base name to be used in saving the plots (default = None)
                flag_show_: boolean setting if the plot gets showed or not (default = True)

        Output:
            Default = None, interactive view
            If flag_save = True then outputs a png image
    """
    if MATPLOTLIB_PRESENT:
        if SCIPY_PRESENT:
            _density_scatter( array1 , array2,  ax = None)
        else:
            plt.scatter(array1, array2)

        if flag_xy_line:
            xmin = np.min(np.append(array1,array2))
            xmax = np.max(np.append(array1,array2))
            plt.plot([xmin,xmax], [xmin,xmax], 'k-', linewidth = 2)

        plt.xlabel(axes_labels[0] + ' ' + "field 1" )
        plt.ylabel(axes_labels[1] + ' ' + "field 2")
        plt.tight_layout()
        if flag_save_:
            if not save_name_:
                base_name = "scatter_"
            else:
                base_name = save_name_+"_"
            save_name = base_name + axes_labels[0] + '_' + axes_labels[1]
            plt.savefig(save_name +'.png')
        if flag_show_:
            plt.show()
    else:
        raise ImportError("matplotlib package not available, please install it for the scatter plot feature")

def matplotlib_simple_distribution(array1, array2, axes_labels,
                                flag_save_ = False, save_name_ = None, flag_show_ = True):
    """ Function to generate an histogram of two data arrays of different length with matplotlib

        Input:
            array1: numpy array of first data set
            array2: numpy array of second data set
            axes_labels: list of string of length 2, used to specify the labels of output axis
            Optional Arguments:
                flag_save_: boolean controlling if the plots get saved (default = False)
                save_name_: (string), base name to be used in saving the plots (default = None)
                flag_show_: boolean setting if the plot gets showed or not (default = True)

        Output:
            Default = None, interactive view
            If flag_save = True then outputs a png image
    """
    import matplotlib.pyplot as plt
    # To Do: look into a "smart" way for bin selection , options include
    #           1) take a fixed nb
    #           2) take longest data array to 'auto' for bin selection
    #           3) take shortest data array to 'auto' for bin selection

    bin_vals, bins, patches = plt.hist(array1, bins='auto',
                                 density=False, facecolor='cornflowerblue', alpha=0.8, label="field 1")
    bin_vals, bins, patches = plt.hist(array2, bins = bins,
                                 density=False, facecolor='lightgreen', alpha=0.6, label="field 2")
    plt.xlabel(axes_labels[0] + ' ')
    plt.ylabel("Count")
    plt.legend()
    plt.tight_layout()
    if flag_save_:
        if not save_name_:
            base_name = "hist_"
        else:
            base_name = save_name_+"_"
        save_name = base_name + axes_labels[0]
        plt.savefig(save_name +'.png')
    if flag_show_:
        plt.show()



"""
    Module controlling YAML and hdf5 output files
"""
import yaml
import h5py
import hdfdict
import os

__all__ = ["write_dict_to_yaml","write_h5"]


def write_dict_to_yaml(dict_, out_name_, setindent = 2):
    """ Generic function writin a dictionary as a yaml file

        Input:
            :dict_: dictionary to write out as yaml
            :out_name_: output name of the .yml file
            :setindent: indentation setting of yaml file
        Output:
            a YAML file
    """
    with open(out_name_+'.yml', 'w') as yaml_out:
        yaml.dump(dict_, yaml_out, indent = setindent)


def write_h5(dict_, save_name_ = 'dump', save_path_='./'):
    tmp_dir = os.getcwd()
    if not os.path.exists(save_path_):
        os.makedirs(save_path_)
    os.chdir(save_path_)
    with h5py.File(save_name_ + ".h5", "w") as fout:
        hdfdict.dump(dict_, fout, lazy=False)
    os.chdir(tmp_dir)
    


""" Command line module of h5cross"""

import click
import h5cross

# not to be used aside of the CLI
__all__ = []

@click.group()
def main_cli():
    """---------------    H5CROSS  --------------------

You are now using the Command line interface of h5cross
a Python3 helper to explore and compare hdf5 files, created at CERFACS (https://cerfacs.fr).


This is a python package currently installed in your python environement.
"""


@click.command()
@click.argument("filename", nargs=1)
def tree(filename):
    """Print the content of an hdf5 file in terminal.

    """
    from h5cross.visit_h5 import print_h5_structure
    
    print_h5_structure(filename)

main_cli.add_command(tree)


# NOTE: QUICK COMMANDS: e.g. @click.option('--out-name', '-n', 
# -n will be shortcut for --out-name
# -i  for out-indent 


@click.command()
@click.argument("filename", nargs=1)
@click.option('--out-name', required=False, default=None, type=str,
    help='default = None , if specified used as YAML output name. \
                            If not specified output name = filename.yml')
@click.option('--out-indent', required=False, default=2, type=int,
    help ='default = 2, sets the indentation level in ouput YAML file')
def dump(filename, out_name, out_indent):
    """Write the content of an hdf5 file into a YAML file.

    """
    from h5cross.view_h5 import get_h5_structure
    mydict = get_h5_structure(filename)

    if out_name is not None:
        writename = out_name
    else:
        writename = filename.split('.')[0]

    configure_yaml_output(mydict,writename, out_indent)

main_cli.add_command(dump)


@click.command()
@click.argument("filename", nargs=1)
def view(filename):
    """Show the content of an hdf5 file interactively with nobvisual.

    """

    from h5cross.view_h5 import visual_h5

    try:
        visual_h5(filename)
    except ImportError as excep:
        print("'h5cross view' stopped:", excep)
main_cli.add_command(view)


@click.command()
@click.argument("filename", nargs=1)
@click.option('--print-to-console', required=False, default=True, type=bool,
    help='default = True')
@click.option('--pretty-table', required=False, default=True, type=bool,
    help='default = True, controls if pretty table output is selected')
@click.option('--pretty-full-path', required=False, default=True, type=bool,
    help='default = True, controls if full paths in nested dictionary are output in pretty table')          
@click.option('--save-as-yaml', required=False, default=False, type=bool,
    help='default = False , if True: default output name is set to input file name')
@click.option('--out-name-yaml', required=False, default=None, type=str,
    help= 'default = None, if specified used as output name of YAML file')
@click.option('--out-indent', required=False, default=2, type=int,
    help ='default = 2, sets the indentation level in ouput YAML file')
# @click.option('--save-as-h5', required=False, default=False, type=bool,
#                     help='To be implemented')
# @click.option('--out-name-h5', required=False, default=None, type=str,
#                     help='To be implemented')
def stats(filename, print_to_console, pretty_table, pretty_full_path, save_as_yaml, out_name_yaml, out_indent):
#            save_as_h5, out_name_h5):
    """Compute statistics of arrays from hdf5 file.

    """
    
    from h5cross.visit_h5 import (get_h5_structure,
                                    get_h5_field,
                                    pretty_table_h5_stats,
                                    print_h5_structure)

    from h5cross.tools import add_numpy_statistics

    mydict = get_h5_structure(filename)
    myfield_dict = get_h5_field(filename)

    add_numpy_statistics(mydict, myfield_dict)
    if print_to_console:
        if pretty_table:
            try:
                pretty_table_h5_stats(mydict, full_path = pretty_full_path)
            except ImportError as excep:
                print("'h5cross stats' stopped:", excep)
        else:
            print_h5_structure(mydict)

    if save_as_yaml:
        if out_name_yaml is not None:
            writename = out_name_yaml
        else:
            writename = filename.split('.')[0]

        configure_yaml_output(mydict, writename, out_indent)

    
    # TO DO: rethink the save_as-> should be in the flow file, so not logic in struct
    #if save_as_h5:
    #    print("Currently not implemented")
    #     if out_name_h5 is not None:
    #         writename = out_name_h5
    #     else:
    #         writename = filename.split('.')[0] + "_stats"

    #     h5cross.write_h5(mydict,save_name = writename)

main_cli.add_command(stats)


@click.command()
@click.argument("file_left", nargs=1)
@click.argument("file_right", nargs=1)
@click.option('--add-stats', required=False, default=False, type=bool,
     help='default = False, if True will compute statistics for each file before comparing')
def diff(file_left, file_right, add_stats):
    """Compare the content of two hdf5 files and view interactively with nobvisual.

    """

    from h5cross.view_h5 import compare_h5

    try:
        compare_h5(file_left, file_right, add_stats)
    except ImportError as excep:
        print("'h5cross diff' stopped:", excep)
        return

main_cli.add_command(diff)

@click.command()
@click.argument("file_left", nargs=1)
@click.argument("file_right", nargs=1)
@click.option('--select-vars', required=True, default=None, type=str,
    help='default = None, select variables to plot. Requires a comma separated string with \
            variables to select, e.g. \"temperature,pressure\" ')
@click.option('--save-output', required=False, default=False, type=bool,
    help='default = False, if True will save the outputs of the plot(s)')
@click.option('--save-name', required=False, default=None, type=str,
    help='default = None, string specifying desired output base name for each plot')
@click.option('--show-output', required=False, default=True, type=bool,
    help='default = True, controls whether plot(s) are shown or not')
@click.option('--use-seaborn', required=False, default=True, type=bool,
    help='default = True, controls the use of seaborn. If False, matplotlib.pyplot is used.')
# @click.option('--add-xy', required=False, default=False, type=bool,
#     help='default = False, If True, superimposes the x=y line on the scatter plot.')
@click.option('--add-xy', required=False, is_flag=True,
    help='flag, if activated: superimposes the x=y line on the scatter plot.')



def scatter(file_left, file_right, select_vars,
            save_output, save_name,
            show_output, use_seaborn,
            add_xy):
    """ Scatter plot comparison of two hdf5 files.
        Seaborn is used to generate the plots but can
        be deactivated in which case matplotlib.pyplot is used.

        Note: The matplotlib package is a minimal requirement for this functionality.

    """
    from h5cross.view_h5 import compare_scatter_h5

    var_list = select_vars.split(",")
    try:
        compare_scatter_h5(file_left, file_right, var_list,
                flag_save=save_output, save_name=save_name,
                flag_show=show_output, flag_seaborn= use_seaborn,
                flag_xy_line = add_xy)
    except ImportError as excep:
        print("'h5cross scatter' stopped:", excep)
        return

main_cli.add_command(scatter)


@click.command()
@click.option('--array-length', required=False, default=None, type=int,
    help='default = None, controls the length of the data arrays')
@click.option('--out-name-h5', required=False, default=None, type=str,
    help='Optional output name for hdf5 file, default = dump.h5')
@click.option('--out-location-h5', required=False, default=None, type=str,
    help='Optional path where to save the hdf5 file, default = current dir')

def generate(array_length, out_name_h5, out_location_h5):
    """ Generates an hdf5 file from a random nested dictionary (also generated).
        It allows an easy input to test the other functionalities.
    """
    from h5cross.generate_h5 import generate_random_nested_dictionary
    from h5cross.write_h5 import write_h5

    # Todo: remove unnecessary elses
    if array_length:
        tmp_dict = generate_random_nested_dictionary(array_length=array_length)
    else:
        tmp_dict = generate_random_nested_dictionary()

    if out_name_h5 is not None:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_name_ = out_name_h5, save_path_= out_location_h5)
        else:
            write_h5(tmp_dict, save_name_ = out_name_h5)
    else:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_path_= out_location_h5)
        else:
            write_h5(tmp_dict)


main_cli.add_command(generate)


@click.command()
@click.argument("filename", nargs=1)
@click.option('--file-type', required=True, type=click.Choice(['nek5000','pvtu','vtu','vtk']), default=None,
    help='default = None, input format to convert to hdf5')
@click.option('--out-name-h5', required=False, default=None, type=str,
    help='Optional output name for hdf5 file, default = dump.h5')
@click.option('--out-location-h5', required=False, default=None, type=str,
    help='Optional path where to save the hdf5 file, default = current dir')

def convert(filename,file_type, out_name_h5, out_location_h5):
    """ Conversion to hdf5 of certain file formats. 
        Currently supported:
            - nek5000: requires "pymech" package
            - vtk: pvtu, vtu   requires "vtk" package
    """

    from h5cross.convert_h5 import convert_nek5000, convert_vtu, convert_vtk
    from h5cross.write_h5 import write_h5

    tmp_dict = dict()
    try:
        if file_type == "nek5000":
            tmp_dict = convert_nek5000(filename)
        elif file_type == "pvtu":
            tmp_dict = convert_vtu(filename, parallel_ = True)
        elif file_type == "vtu":
            tmp_dict = convert_vtu(filename, parallel_ = False)
        elif file_type == "vtk":
            tmp_dict = convert_vtk(filename)
    except ImportError as excep:
        print("'h5cross convert' stopped:", excep)
        return
    except AssertionError as excep:
        print("'h5cross convert' stopped:", excep)
        return
    except ValueError as excep:
        print("'h5cross convert' stopped:", excep)
        return

    if out_name_h5 is not None:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_name_ = out_name_h5, save_path_ = out_location_h5)
        else:
            write_h5(tmp_dict, save_name_ = out_name_h5)
    else:
        if out_location_h5 is not None:
            write_h5(tmp_dict, save_path_= out_location_h5)
        else:
            write_h5(tmp_dict)


main_cli.add_command(convert)


#----------------------extra functionalities------------------------#
def configure_yaml_output(dict_, out_name_, indent_):
    """ Function to set output options and write the yaml output

        Input:
            :dict_: dictionary to write out as a yaml file
            :out_name_: dtype = string, name of output yaml file
            :indent_: dtype = int, indentation level for output
        Output:
            writes a yaml file in current directory
    """

    from h5cross.write_h5 import write_dict_to_yaml
    if indent_ is None:
        indent_ = 2
    write_dict_to_yaml(dict_, out_name_, setindent=indent_)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

"""
    h5cross

"""
import os.path as path
from glob import glob
from os.path import basename
#from os.path import dirname
from os.path import join
from os.path import splitext
from setuptools import find_packages, setup


# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
     long_description = f.read()

with open(path.join(this_directory, 'VERSION'), encoding='utf-8') as f:
    version = f.read()
    print(version)

with open('LICENSE.txt') as f:
     license = f.read()

setup(
    name= "h5cross",
    version=version,
    description="comparison of hdf5 files",
    author = "Jimmy-John Hoste",
    author_email="coop@cerfacs.fr",
    url="",
    license ="MIT License",
    license_files = license,
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["hdf5","data comparison", "visualisation"],
     classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        "setuptools",
        "PyYAML",
        "nobvisual>=0.2.0",
        "h5py",
        "hdfdict>=0.3.1",
        "numpy",
        "click"
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "h5cross = h5cross.cli:main_cli",
        ]
    },
)



"""   Functionalities related to the navigation of hdf5 type files
"""
import sys
import h5py
import numpy as np
import yaml
import hdfdict

from .tools import indentify_next_nested_level

PRETTYTABLE_PRESENT = True
try:
    from prettytable import PrettyTable
except ImportError:
    PRETTYTABLE_PRESENT = False

__all__ = ["get_h5_structure", "print_h5_structure","get_h5_field","pretty_table_h5_stats"]


def get_h5_structure(h5_filename):
    """ Show hdf5 file components

        Input:
            :h5_filename: path to hdf5 file to inspect
    """

    with h5py.File(h5_filename, "r") as node:
        out = log_hdf_node(node)
    return out

def add_nested_items_to_table(dict_,table_, item_list_, complete_path_ = False, extra_arg_=""):
    """ Function to navigate through a nested object and add desired keys
        to a PrettyTable.

        Input:
            :dict_ : input dictionary to navigate through
            :table_: PrettyTable to which to add desired items
            :item_list_: list of type string with dict items to add to table_
            :complete_path_: boolean to select if all nested levels are added to item
                            (default = False)
            :extra_arg_: variable of type string used if complete_path_ = True.
                        Required in that case because of recursive calls.
        Output:
            :None: add rows to input table_
    """

    for key in dict_.keys():
        # check if we are at most nested level
        dict_type_ = indentify_next_nested_level(dict_[key])
        tmp_key = list(dict_type_.keys())[0]
        if not dict_type_[tmp_key]:
            if complete_path_:
                items_add = [extra_arg_+"/"+key]
            else:
                items_add = [key]

            for items in item_list_:
                if items in dict_[key].keys():
                    items_add.append(np.round(float(dict_[key][items]),8))
                else:
                    items_add.append("NA")
            table_.add_row(items_add)
        else:
            # we are not at most nested level, recursive call to next nested level
            extra_arg_ += "/"+key
            add_nested_items_to_table(dict_[key], table_, item_list_,
                                complete_path_=complete_path_,extra_arg_=extra_arg_)
            extra_arg_ = ""


def pretty_table_h5_stats(dict_, full_path = True):
    """ Function to generate a pretty table output of the added statistics
        min, mean, max, std and median.

        Input:
            :dict_: dictionary with added statistics
            :full_path: boolean, if True, full paths outputs are used for the nested
                    dict_ input (default = True)
        Output:
            :None: print to console
    """

    if PRETTYTABLE_PRESENT:
        # Could render this optional in future
        table = PrettyTable(['Dataset','min','mean','max','st dev','median'])
        item_list = ['min','mean','max','std','median']
        add_nested_items_to_table(dict_,table, item_list, complete_path_= full_path)
        table.align['Dataset'] = "l"
        print(table)
    else:
        raise ImportError("prettytable package not available, please install it for this feature or deactivate its use")


def print_h5_structure(h5_filename):
    """ Show hdf5 file components

        Input:
            :h5_filename: path to hdf5 file to inspect
        Output:
            Print to terminal
    """

    if isinstance(h5_filename,str):
        dict_ = get_h5_structure(h5_filename)
    else:
        dict_ = h5_filename
        # does weird stuff if dict_ contains data.
        # Must be a structure dictionary
    print(yaml.dump(dict_, default_flow_style=False))


def get_h5_field(h5_filename):
    """ Return a dictionary of an hdf5 file, e.g. a
        flow field or mesh file.

        Input:
            :h5_filename: path to the file to be read in
        Output:
            :out_: field dictionary
    """

    with h5py.File(h5_filename, "r") as h5pf:
        out_ = hdfdict.load(h5pf, lazy=False)
    return out_


def log_hdf_node(node):
    """
    Build a dictionnary with the structure of a HDF5 node

    Parameters:
    -----------
    node : hdf5 node

    Returns:
    --------
    a dictionnary
    """

    out = dict()


    def extend_dict(dict_, address, attr):
        tmp = dict_
        #print(attr)
        #!! Needs to be adapted for a multi-level nested object!!
        # Single level works fine, because we currently add attr to each level
        #  of the address, should be on the last level only -> 2 loops?
        for key in address[:]:
            if key not in tmp:
                #tmp[key] = dict()
                #print("Items current key and last key:", key, address[-1])
                if key == address[-1]:
                    #print("Last key", tmp)
                    tmp[key] = attr.copy()
                else:
                    tmp[key] = {}
            #print(tmp)
            tmp = tmp[key] # so we redefine the dict to continue through

    def visitor_func(name, node):
        key_list = [item.strip() for item in name.split('/')]
        #print(key_list)
        if isinstance(node, h5py.Dataset):
            #print("Create dict with dtype and value")
            attr = dict()
            attr["dtype"] = str(node.dtype)
            attr["value"] = _get_node_description(node)
            #print(out)
            extend_dict(out, key_list, attr)
            #print(out)
        else:
            pass

    node.visititems(visitor_func)

    return out

def _ascii2string(ascii_list):
    """ Ascii to string conversion

        Parameters:
        -----------
        ascii_list : a list of string to be converted

        Returns:
        --------
        a string joining the list elements

    """

    return ''.join(chr(i) for i in ascii_list[:-1])


def _get_node_description(node):
    """Get number of elements in an array or
      value of a single-valued node.

    Parameters:
    -----------
    node : hdf5 node

    Returns:
    --------
    a value with a Python format
    None if data is not a singlevalued quantity
    """

    out = None
    value = node[()]
    shape = node.shape
    #print(node)
    #print(shape)
    if np.prod(shape) == 1:
        # this is a strong assumption because if you find int8
        # your are probably looking at an hdf5 file applying the cgns standard
        #print(node)
        if node.dtype in ["int8"]:
            #print(_ascii2string(value))
            out = np.char.array(_ascii2string(value))[0]
        elif shape in [(1), (1,)]:
            if node.dtype in ["int32", "int64"]:
                out = int(value[0])
            elif node.dtype in ["float32", "float64"]:
                out = float(value[0])
            #else:
            #    raise RuntimeError()
        # else:
        #     raise RuntimeError()
    else:
        out = "array of %s elements" %(" x ".join([str(k) for k in shape]))
    return out
